@@ -23,6 +23,8 @@
 #include "../shumate-vector-sprite-sheet.h"
 #include "shumate-vector-value-private.h"
 #include "shumate-vector-utils-private.h"
+#include "../shumate-vector-reader-iter-private.h"
+#include "shumate-vector-index-private.h"
 
 typedef enum {
   SHUMATE_VECTOR_GEOMETRY_POINT = VECTOR_TILE__TILE__GEOM_TYPE__POINT,
@@ -39,6 +41,7 @@ typedef struct {
   int tile_x;
   int tile_y;
   int layer_idx;
+  int source_layer_idx;
 
   GPtrArray *symbols;
 
@@ -46,13 +49,12 @@ typedef struct {
 
   float overzoom_x, overzoom_y, overzoom_scale;
 
-  VectorTile__Tile *tile;
-  VectorTile__Tile__Layer *layer;
-  VectorTile__Tile__Feature *feature;
+  ShumateVectorReaderIter *reader;
+  ShumateVectorIndex *index;
+  ShumateVectorIndexDescription *index_description;
 } ShumateVectorRenderScope;
 
 
-gboolean shumate_vector_render_scope_find_layer (ShumateVectorRenderScope *self, const char *layer_name);
 void shumate_vector_render_scope_exec_geometry (ShumateVectorRenderScope *self);
 void shumate_vector_render_scope_get_geometry_center (ShumateVectorRenderScope *self, double *x, double *y);
 void shumate_vector_render_scope_get_bounds (ShumateVectorRenderScope *self,
@@ -68,3 +70,4 @@ void shumate_vector_render_scope_get_variable (ShumateVectorRenderScope *self, c
 
 GHashTable *shumate_vector_render_scope_create_tag_table (ShumateVectorRenderScope *self);
 
+void shumate_vector_render_scope_index_layer (ShumateVectorRenderScope *self);
